#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from dataclasses import dataclass
from typing import Any

from .file import FILE, read_bson, write_bson


@dataclass(frozen=True)
class Data:
    def to_bson(self, file: FILE) -> int:
        return write_bson(file, self.__dict__)

    @classmethod
    def from_bson(cls, file: FILE) -> Any:
        x = read_bson(file)
        return cls(**x)

    @classmethod
    def keys(cls):
        yield from cls.__dict__["__annotations__"].keys()

    def items(self):
        yield from self.__dict__.items()



from dataclasses import dataclass

import numpy as np
import pandas as pd
import polars as pl

from cvx.bson.dataclass import Data


@dataclass(frozen=True)
class Maffay(Data):
    x: pd.DataFrame
    y: pl.DataFrame
    z: np.array


@dataclass(frozen=True)
class DataAPI(Data):
    # you need to explicitly declare the tables expected
    x: pd.DataFrame
    y: pl.DataFrame
    z: np.array

    def items(self):
        yield from self.__dict__.items()


def assert_equal(obj1, obj2):
    assert type(obj1) is type(obj2)

    if isinstance(obj1, pd.DataFrame):
        pd.testing.assert_frame_equal(obj1, obj2)

    if isinstance(obj1, np.ndarray):
        np.testing.assert_array_equal(obj1, obj2)

    if isinstance(obj1, pl.DataFrame):
        assert obj1.equals(obj2)


def test_conversion(tmp_path):
    matrix = np.random.rand(5, 2)

    x = pd.DataFrame(data=matrix)
    y = pl.DataFrame(data=2 * matrix)
    z = matrix

    data = Maffay(x=x, y=y, z=z)

    print(data.to_bson(file=tmp_path / "test.bson"))
    print(Maffay.from_bson(file=tmp_path / "test.bson"))


def test_reflection(tmp_path):
    matrix = np.random.rand(5, 2)

    x = pd.DataFrame(data=matrix)
    y = pl.DataFrame(data=2 * matrix)
    z = matrix

    data = {"x": x, "y": y, "z": z, "data": "xxx"}
    data = {key: value for key, value in data.items() if key in DataAPI.keys()}
    api = DataAPI(**data)

    for key, value in api.items():
        assert_equal(data[key], value)



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json
from io import BytesIO
from typing import Any, Union

import numpy as np
import pandas as pd
import polars as pl
import pyarrow as pa


def encode(data: Union[np.ndarray, pd.DataFrame, pl.DataFrame]) -> Any:
    """
    Encode a numpy array or a pandas DataFrame

    Args:
        data: The numpy array or pandas DataFrame

    Returns: object converted into bytes
    """
    if isinstance(data, np.ndarray):
        tensor = pa.Tensor.from_numpy(obj=data)
        buffer = pa.BufferOutputStream()
        pa.ipc.write_tensor(tensor, buffer)
        return bytes(buffer.getvalue().to_pybytes())

    if isinstance(data, pd.DataFrame):
        return data.to_parquet()

    if isinstance(data, pl.DataFrame):
        result = data.write_ipc(file=None)
        result.seek(0)
        return result.read()

    converted = json.dumps(data).encode(encoding="utf-8")
    arr = bytes("cvx", "utf-8")
    return arr + converted

    # return bytes.
    # print(encoded_tuple)
    # decoded_color = encoded_color.decode()
    # original_form = json.load(decoded_color)
    # return

    # raise TypeError(f"Invalid Datatype {type(data)}")


def decode(data: bytes) -> Union[np.ndarray, pd.DataFrame, pl.DataFrame]:
    """
    Decode the bytes back into numpy array or pandas DataFrame

    Args:
        data: bytes

    Returns:
        The array or the frame
    """
    # reader the first few bytes
    header = data[:3]

    # ARR indicates a pl.DataFrame
    if header == b"ARR":
        return pl.read_ipc(data)

    # PAR indicates a pd.DataFrame
    if header == b"PAR":
        return pd.read_parquet(BytesIO(data))

    if header == b"cvx":
        return json.loads(data[3:].decode())

    # if still here we try numpy
    return pa.ipc.read_tensor(data).to_numpy()



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json

import numpy as np


class NumpyEncoder(json.JSONEncoder):
    """Custom encoder for numpy data types"""

    def default(self, obj):
        if isinstance(
            obj,
            (
                np.int_,
                np.intc,
                np.intp,
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ),
        ):
            return int(obj)

        elif isinstance(obj, (np.float16, np.float32, np.float64)):
            return float(obj)

        elif isinstance(obj, (np.complex64, np.complex128)):
            return {"real": obj.real, "imag": obj.imag}

        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()

        elif isinstance(obj, np.bool_):
            return bool(obj)

        elif isinstance(obj, np.void):
            return None

        return json.JSONEncoder.default(self, obj)
